"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.SlidingBuffer = exports.DroppingBuffer = exports.FixedBuffer = exports.RingBuffer = undefined;

var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require("babel-runtime/helpers/createClass");

var _createClass3 = _interopRequireDefault(_createClass2);

exports.ring = ring;
exports.fixed = fixed;
exports.dropping = dropping;
exports.sliding = sliding;

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var RingBuffer = exports.RingBuffer = function () {
  function RingBuffer() {
    (0, _classCallCheck3.default)(this, RingBuffer);

    this._array = [];
  }

  (0, _createClass3.default)(RingBuffer, [{
    key: "unshift",
    value: function unshift(item) {
      this._array.unshift(item);
    }
  }, {
    key: "pop",
    value: function pop() {
      return this._array.pop();
    }
  }, {
    key: "count",
    value: function count() {
      return this._array.length;
    }
  }, {
    key: "cleanup",
    value: function cleanup(predicate) {
      this._array = this._array.filter(predicate);
    }
  }]);
  return RingBuffer;
}();

var FixedBuffer = exports.FixedBuffer = function () {
  function FixedBuffer(buffer, n) {
    (0, _classCallCheck3.default)(this, FixedBuffer);

    this._buffer = buffer;
    this._n = n;
  }

  (0, _createClass3.default)(FixedBuffer, [{
    key: "isFull",
    value: function isFull() {
      return this._buffer.count() >= this._n;
    }
  }, {
    key: "add",
    value: function add(item) {
      this._buffer.unshift(item);
    }
  }, {
    key: "remove",
    value: function remove() {
      return this._buffer.pop();
    }
  }, {
    key: "count",
    value: function count() {
      return this._buffer.count();
    }
  }]);
  return FixedBuffer;
}();

var DroppingBuffer = exports.DroppingBuffer = function () {
  function DroppingBuffer(buffer, n) {
    (0, _classCallCheck3.default)(this, DroppingBuffer);

    this._buffer = buffer;
    this._n = n;
  }

  (0, _createClass3.default)(DroppingBuffer, [{
    key: "isFull",
    value: function isFull() {
      return false;
    }
  }, {
    key: "add",
    value: function add(item) {
      if (this._buffer.count() < this._n) {
        this._buffer.unshift(item);
      }
    }
  }, {
    key: "remove",
    value: function remove() {
      return this._buffer.pop();
    }
  }, {
    key: "count",
    value: function count() {
      return this._buffer.count();
    }
  }]);
  return DroppingBuffer;
}();

var SlidingBuffer = exports.SlidingBuffer = function () {
  function SlidingBuffer(buffer, n) {
    (0, _classCallCheck3.default)(this, SlidingBuffer);

    this._buffer = buffer;
    this._n = n;
  }

  (0, _createClass3.default)(SlidingBuffer, [{
    key: "isFull",
    value: function isFull() {
      return false;
    }
  }, {
    key: "add",
    value: function add(item) {
      if (this._buffer.count() === this._n) {
        this._buffer.pop();
      }

      this._buffer.unshift(item);
    }
  }, {
    key: "remove",
    value: function remove() {
      return this._buffer.pop();
    }
  }, {
    key: "count",
    value: function count() {
      return this._buffer.count();
    }
  }]);
  return SlidingBuffer;
}();

/**
 * Returns a buffer that is considered "full" when it reaches size n,
 * but still accepts additional items, effectively allow overflowing.
 * The overflowing behavior is useful for supporting "expanding"
 * transducers, where we want to check if a buffer is full before
 * running the transduced step function, while still allowing a
 * transduced step to expand into multiple "essence" steps.
 */


function ring() {
  return new RingBuffer();
}

function fixed(n) {
  return new FixedBuffer(ring(), n);
}

function dropping(n) {
  return new DroppingBuffer(ring(), n);
}

function sliding(n) {
  return new SlidingBuffer(ring(), n);
}