'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Channel = exports.PutBox = exports.Box = exports.CLOSED = exports.MAX_QUEUE_SIZE = exports.MAX_DIRTY = undefined;

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

exports.chan = chan;

var _dispatch = require('./dispatch');

var _buffers = require('./buffers');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var MAX_DIRTY = exports.MAX_DIRTY = 64;
var MAX_QUEUE_SIZE = exports.MAX_QUEUE_SIZE = 1024;
var CLOSED = exports.CLOSED = null;

var isReduced = function isReduced(v) {
  return v && v['@@transducer/reduced'];
};
var schedule = function schedule(f, v) {
  return (0, _dispatch.run)(function () {
    return f(v);
  });
};

var Box = exports.Box = function Box(value) {
  (0, _classCallCheck3.default)(this, Box);

  this.value = value;
};

var PutBox = exports.PutBox = function PutBox(handler, value) {
  (0, _classCallCheck3.default)(this, PutBox);

  this.handler = handler;
  this.value = value;
};

var Channel = exports.Channel = function () {
  function Channel(takes, puts, buf, xform) {
    (0, _classCallCheck3.default)(this, Channel);

    this.buf = buf;
    this.xform = xform;
    this.takes = takes;
    this.puts = puts;
    this.dirtyTakes = 0;
    this.dirtyPuts = 0;
    this.closed = false;
  }

  (0, _createClass3.default)(Channel, [{
    key: 'put',
    value: function put(value, handler) {
      if (value === CLOSED) {
        throw new Error('Cannot put CLOSED on a channel.');
      }

      // TODO: I'm not sure how this can happen, because the operations
      // are registered in 1 tick, and the only way for this to be inactive
      // is for a previous operation in the same alt to have returned
      // immediately, which would have short-circuited to prevent this to
      // be ever register anyway. The same thing goes for the active check
      // in "take".
      if (!handler.isActive()) {
        return null;
      }

      if (this.closed) {
        handler.commit();
        return new Box(false);
      }

      // Soak the value through the buffer first, even if there is a
      // pending taker. This way the step function has a chance to act on the
      // value.
      if (this.buf && !this.buf.isFull()) {
        handler.commit();
        var done = isReduced(this.xform['@@transducer/step'](this.buf, value));

        // flow-ignore
        while (this.takes.count() > 0 && this.buf.count() > 0) {
          var taker = this.takes.pop();

          if (taker.isActive()) {
            // flow-ignore
            schedule(taker.commit(), this.buf.remove());
          }
        }
        if (done) {
          this.close();
        }
        return new Box(true);
      }

      // Either the buffer is full, in which case there won't be any
      // pending takes, or we don't have a buffer, in which case this loop
      // fulfills the first of them that is active (note that we don't
      // have to worry about transducers here since we require a buffer
      // for that).
      while (this.takes.count() > 0) {
        var _taker = this.takes.pop();

        if (_taker.isActive()) {
          handler.commit();
          var callback = _taker.commit();
          schedule(callback, value);
          return new Box(true);
        }
      }

      // No buffer, full buffer, no pending takes. Queue this put now if blockable.
      if (this.dirtyPuts > MAX_DIRTY) {
        this.puts.cleanup(function (putter) {
          return putter.handler.isActive();
        });
        this.dirtyPuts = 0;
      } else {
        this.dirtyPuts++;
      }

      if (handler.isBlockable()) {
        if (this.puts.count() >= MAX_QUEUE_SIZE) {
          throw new Error('No more than ' + MAX_QUEUE_SIZE + ' pending puts are allowed on a single channel.');
        }
        this.puts.unshift(new PutBox(handler, value));
      }

      return null;
    }
  }, {
    key: 'take',
    value: function take(handler) {
      if (!handler.isActive()) {
        return null;
      }

      if (this.buf && this.buf.count() > 0) {
        handler.commit();
        // flow-ignore
        var value = this.buf.remove();

        // We need to check pending puts here, other wise they won't
        // be able to proceed until their number reaches MAX_DIRTY

        // flow-ignore
        while (this.puts.count() > 0 && !this.buf.isFull()) {
          // flow-ignore
          var putter = this.puts.pop();

          if (putter.handler.isActive()) {
            var callback = putter.handler.commit();

            if (callback) {
              schedule(callback, true);
            }

            if (isReduced(this.xform['@@transducer/step'](this.buf, putter.value))) {
              this.close();
            }
          }
        }
        return new Box(value);
      }

      // Either the buffer is empty, in which case there won't be any
      // pending puts, or we don't have a buffer, in which case this loop
      // fulfills the first of them that is active (note that we don't
      // have to worry about transducers here since we require a buffer
      // for that).
      while (this.puts.count() > 0) {
        // flow-ignore
        var _putter = this.puts.pop();

        if (_putter.handler.isActive()) {
          handler.commit();
          var _callback = _putter.handler.commit();

          if (_callback) {
            schedule(_callback, true);
          }

          return new Box(_putter.value);
        }
      }

      if (this.closed) {
        handler.commit();
        return new Box(CLOSED);
      }

      // No buffer, empty buffer, no pending puts. Queue this take now if blockable.
      if (this.dirtyTakes > MAX_DIRTY) {
        this.takes.cleanup(function (_handler) {
          return _handler.isActive();
        });
        this.dirtyTakes = 0;
      } else {
        this.dirtyTakes++;
      }

      if (handler.isBlockable()) {
        if (this.takes.count() >= MAX_QUEUE_SIZE) {
          throw new Error('No more than ' + MAX_QUEUE_SIZE + ' pending takes are allowed on a single channel.');
        }

        this.takes.unshift(handler);
      }

      return null;
    }
  }, {
    key: 'close',
    value: function close() {
      if (this.closed) {
        return;
      }

      this.closed = true;

      // TODO: Duplicate code. Make a "_flush" function or something
      if (this.buf) {
        this.xform['@@transducer/result'](this.buf);

        // flow-ignore
        while (this.takes.count() > 0 && this.buf.count() > 0) {
          var taker = this.takes.pop();

          if (taker.isActive()) {
            // flow-ignore
            schedule(taker.commit(), this.buf.remove());
          }
        }
      }

      while (this.takes.count() > 0) {
        var _taker2 = this.takes.pop();

        if (_taker2.isActive()) {
          schedule(_taker2.commit(), CLOSED);
        }
      }

      while (this.puts.count() > 0) {
        // flow-ignore
        var putter = this.puts.pop();

        if (putter.handler.isActive()) {
          var pulCallback = putter.handler.commit();

          if (pulCallback) {
            schedule(pulCallback, false);
          }
        }
      }
    }
  }, {
    key: 'isClosed',
    value: function isClosed() {
      return this.closed;
    }
  }]);
  return Channel;
}();

// The base transformer object to use with transducers


var AddTransformer = {
  '@@transducer/init': function transducerInit() {
    throw new Error('init not available');
  },

  '@@transducer/result': function transducerResult(v) {
    return v;
  },

  '@@transducer/step': function transducerStep(buffer, input) {
    buffer.add(input);
    return buffer;
  }
};

function defaultExceptionHandler(err) {
  console.log('error in channel transformer', err.stack); // eslint-disable-line
  return CLOSED;
}

function handleEx(buf, exHandler, e) {
  var def = (exHandler || defaultExceptionHandler)(e);

  if (def !== CLOSED) {
    // flow-ignore
    buf.add(def);
  }

  return buf;
}

function handleException(exHandler) {
  return function (xform) {
    return {
      '@@transducer/step': function transducerStep(buffer, input) {
        try {
          return xform['@@transducer/step'](buffer, input);
        } catch (e) {
          return handleEx(buffer, exHandler, e);
        }
      },
      '@@transducer/result': function transducerResult(buffer) {
        try {
          return xform['@@transducer/result'](buffer);
        } catch (e) {
          return handleEx(buffer, exHandler, e);
        }
      }
    };
  };
}

// XXX: This is inconsistent. We should either call the reducing
// function xform, or call the transducer xform, not both
function chan(buf, xform, exHandler) {
  var newXForm = void 0;

  if (xform) {
    if (!buf) {
      throw new Error('Only buffered channels can use transducers');
    }

    newXForm = xform(AddTransformer);
  } else {
    newXForm = AddTransformer;
  }

  return new Channel((0, _buffers.ring)(), (0, _buffers.ring)(), buf, handleException(exHandler)(newXForm));
}