'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Process = exports.NO_VALUE = undefined;

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

exports.putThenCallback = putThenCallback;
exports.takeThenCallback = takeThenCallback;
exports.take = take;
exports.put = put;
exports.sleep = sleep;
exports.alts = alts;
exports.poll = poll;
exports.offer = offer;

var _select = require('./select');

var _handlers = require('./handlers');

var _instruction = require('./instruction');

var _instruction2 = _interopRequireDefault(_instruction);

var _channels = require('./channels');

var _dispatch = require('./dispatch');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var NO_VALUE = exports.NO_VALUE = '@@process/NO_VALUE';
function putThenCallback(channel, value, callback) {
  var result = channel.put(value, new _handlers.FnHandler(true, callback));

  if (result && callback) {
    callback(result.value);
  }
}

function takeThenCallback(channel, callback) {
  var result = channel.take(new _handlers.FnHandler(true, callback));

  if (result && callback) {
    callback(result.value);
  }
}

function take(channel) {
  return new _instruction2.default(_instruction2.default.TAKE, channel);
}

function put(channel, value) {
  return new _instruction2.default(_instruction2.default.PUT, {
    channel: channel,
    value: value
  });
}

function sleep(msecs) {
  return new _instruction2.default(_instruction2.default.SLEEP, msecs);
}

function alts(operations, options) {
  return new _instruction2.default(_instruction2.default.ALTS, {
    operations: operations,
    options: options
  });
}

function poll(channel) {
  if (channel.closed) {
    return NO_VALUE;
  }

  var result = channel.take(new _handlers.FnHandler(false));

  return result ? result.value : NO_VALUE;
}

function offer(channel, value) {
  if (channel.closed) {
    return false;
  }

  var result = channel.put(value, new _handlers.FnHandler(false));

  return result instanceof _channels.Box;
}

var Process = exports.Process = function () {
  function Process(gen, onFinishFunc, creatorFunc) {
    (0, _classCallCheck3.default)(this, Process);

    this.gen = gen;
    this.creatorFunc = creatorFunc;
    this.onFinishFunc = onFinishFunc;
    this.finished = false;
  }

  // TODO FIX XXX: This is a (probably) temporary hack to avoid blowing
  // up the stack, but it means double queueing when the value is not
  // immediately available


  (0, _createClass3.default)(Process, [{
    key: '_continue',
    value: function _continue(response) {
      var _this = this;

      (0, _dispatch.run)(function () {
        return _this.run(response);
      });
    }
  }, {
    key: '_done',
    value: function _done(value) {
      var _this2 = this;

      if (!this.finished) {
        this.finished = true;
        (0, _dispatch.run)(function () {
          return _this2.onFinishFunc(value);
        });
      }
    }
  }, {
    key: 'run',
    value: function run(response) {
      var _this3 = this;

      if (this.finished) {
        return;
      }

      // TODO: Shouldn't we (optionally) stop error propagation here (and
      // signal the error through a channel or something)? Otherwise the
      // uncaught exception will crash some runtimes (e.g. Node)
      var iter = this.gen.next(response);
      var ins = iter.value;

      if (iter.done) {
        this._done(ins);
      } else if (ins instanceof _instruction2.default) {
        switch (ins.op) {
          case _instruction2.default.TAKE:
            {
              takeThenCallback(ins.data, function (value) {
                return _this3._continue(value);
              });
              break;
            }

          case _instruction2.default.PUT:
            {
              putThenCallback(ins.data.channel, ins.data.value, function (ok) {
                return _this3._continue(ok);
              });
              break;
            }

          case _instruction2.default.SLEEP:
            {
              (0, _dispatch.queueDelay)(function () {
                return _this3.run(null);
              }, ins.data);
              break;
            }

          case _instruction2.default.ALTS:
            {
              (0, _select.doAlts)(ins.data.operations, function (result) {
                return _this3._continue(result);
              }, ins.data.options);
              break;
            }

          default:
            throw new Error('Unhandled instruction: ' + ins.toString());
        }
      } else if (ins instanceof _channels.Channel) {
        takeThenCallback(ins, function (value) {
          return _this3._continue(value);
        });
      } else {
        this._continue(ins);
      }
    }
  }]);
  return Process;
}();